import os
import sys
import reflex
import numpy as np
import copy
import importlib
from astropy.io import fits

try :
    am = importlib.import_module('auto_molecule')
    iu = importlib.import_module('inst_utils')
except ImportError:
    sys.path.insert(0, '%s/instruments' %(os.path.dirname(__file__)))
    am = importlib.import_module('auto_molecule')
    iu = importlib.import_module('inst_utils')
except Exception as e:
    raise(e)

# ------------------------------------------------------------------------------------------
def set_inst_setup(header=None) :
    if header :
      return header.get('HIERARCH ESO INS MODE', 'UNKNOWN')
    return 'None'
# ------------------------------------------------------------------------------------------
def check_format(files, output_dir):
    """
    Create a molecfit compatible input file from an incompatible file:
    e.g. extract a 1-d spectrum from a 2-d spectrum (for example by extracting the fibre with
    the highest S/N (GIRAFFE))

    And, because molecfit works in TOPOCENTRIC+VACUUM (TOPO_VAC == TV), create a
    binary table format version of the input science with WAVE_TV, this means that
    WAVE_INLCUDEs WLs will be consistent for all data, instead of having to be adjusted
    for each data-set depedning on RV (e.g. for ESPRESSO...)
    """

    '''
    ESPRESSO S1D_FINAL_A files should be in VAC_RV == BARYCENTRIC-VACUUM
    And all are IDP compliant, so we just need to add a WAVE_TV column to the existing
    Binary Table
    '''
    output_files=list()
    files_dict={}
    for file in files:
        if file.category not in files_dict.keys() :
            files_dict[file.category]=[]
        files_dict[file.category]+=[file,]
    has_sci_and_std=(
        ('SCIENCE' in files_dict.keys() or 'SCIENCE_CALCTRANS' in files_dict.keys())
        and
        'STD_MODEL' in files_dict.keys()
    )
    for k in files_dict.keys() :
        for file in files_dict[k] :
            if k in ['STD_MODEL','SCIENCE','SCIENCE_CALCTRANS',] :
                hdulist=fits.open(file.name)
                procatg=hdulist[0].header.get('HIERARCH ESO PRO CATG','NULL')
                out_file_name="%s/%s" %(output_dir,os.path.basename(file.name))
                if procatg in [
                    "S1D_A",
                    "S1D_FINAL_A",
                    "S1D_FLUXCAL_A",
                    "S1D_SKYSUB_A",
                    "S1D_SKYSUB_FLUXCAL_A",
                    "S1D_STD_A",
                ] :
                    if "HIERARCH ESO QC BERV" in hdulist[0].header :
                        QC_BERV_VALUE=hdulist[0].header["HIERARCH ESO QC BERV"]
                        clam='wavelength'
                        # Column names depend on the product type...
                        if hdulist[0].header.get('HIERARCH ESO PRO CATG','NULL') in [
                                "S1D_FINAL_A",
                            ] :
                            clam='WAVE'
                        wave_tv=hdulist[1].columns[clam].copy()
                        wave_tv.name='WAVE_TV'
                        wave_tv.array=iu.bary_to_topo(
                            hdulist[1].data[clam],
                            QC_BERV_VALUE,
                        )
                        new_columns=hdulist[1].columns + fits.ColDefs([wave_tv,])
                        nhdu=fits.BinTableHDU.from_columns(new_columns,)
                        nhdu.header+=hdulist[1].header
                        nhdu.header['HIERARCH ESO MF WAVE COLUMN']='WAVE_TV'
                        nhdu.header['HIERARCH ESO MF WAVE COLUMN WAVELENGTH_FRAME']='TOPO+VAC'
                        hdulist[1]=nhdu
                        iu.HDUList(hdulist).writeto(
                            out_file_name, overwrite=True, checksum=True, output_verify='silentfix',
                        )
                        # broadcast the modified file with WAVE_TV column...
                        output_files.append(reflex.FitsFile(
                            out_file_name, file.category, None, file.purposes
                        ))
                        # broadcast the original frame (for sticking it all back together at the end...)
                        output_files.append(reflex.FitsFile(
                            file.name, "ORIG_%s" %(file.category), None, file.purposes
                        ))
            else :
                # broadcast it...
                output_files.append(reflex.FitsFile(file.name, file.category, None, file.purposes ))
    return output_files
# ------------------------------------------------------------------------------------------
wlg = 0.0001
def set_parameters(header=None, hdu=None, previous_inst_setup='None', files=None) :
    wframe='AIR'
    list_mol='NULL'
    fit_mol='NULL'
    rel_col='NULL'
    clam='wavelength'
    cflux='flux'
    cdflux='NULL'
    cmask='NULL'
    fit_tel_back="False"
    cont_poly = int(4)
    varkern="FALSE"
    wave_include='NULL'

    return {
        'WLG_TO_MICRON': wlg,
        'WAVELENGTH_FRAME': wframe,
        'LIST_MOLEC': list_mol,
        'FIT_MOLEC': fit_mol,
        'REL_COL': rel_col,
        'COLUMN_LAMBDA': clam,
        'COLUMN_FLUX': cflux,
        'COLUMN_DFLUX': cdflux,
        'COLUMN_MASK': cmask,
        'FIT_TELESCOPE_BACKGROUND': fit_tel_back,
        'CONTINUUM_N': cont_poly,
        'VARKERN': varkern,
        'WAVE_INCLUDE':wave_include,
    }
# ------------------------------------------------------------------------------------------
def check_idp(hdu, previous_inst_setup='None', files=None, ParameterInitialization=False, recipe=None, ) :
    import numpy as np
    no_output=0
    clam='wavelength'
    cflux='flux'
    cdflux='error'
    #cmask='quality'
    flux_unit=2
    wave_frame="VAC"

    ## ESPRESSO identifies the UT in the ESO.TEL keywords i.e. ESO.TEL<i>...
    ## But the ESO.TELi.TH.M1.TEMP kw doesn't (yet) fit in the model --> calctrans header (too long by 1 char)
    ## So check if the user hasn't set TEL<i> --> Tel
    tel_ids={
        'ESO-VLT-U1': "TEL1",
        'ESO-VLT-U2': "TEL2",
        'ESO-VLT-U3': "TEL3",
        'ESO-VLT-U4': "TEL4",
    }
    tel_id=tel_ids.get(hdu[0].header.get('TELESCOP','NONE'),'TEL')
    tel_kwds={
        "ELEVATION_KEYWORD": "GEOELEV",
        "LATITUDE_KEYWORD": "GEOLAT",
        "LONGITUDE_KEYWORD": "GEOLON",
        "MIRROR_TEMPERATURE_KEYWORD": "TH M1 TEMP",
        "PRESSURE_KEYWORD": "AMBI PRES START",
        "RELATIVE_HUMIDITY_KEYWORD": "AMBI RHUM",
        "TELESCOPE_ANGLE_KEYWORD": "ALT",
        "TEMPERATURE_KEYWORD": "AMBI TEMP",
    }
    tel_vals={}
    if tel_id not in tel_ids :
        # Assume we have multi-UT observation, i.e. TELECSOP='ESO-VLT-U1234' or similar
        # We thus set the KEYWORDs params all to NONE and the VALUE params
        # to the median of the N-values
        print("Frame not recognised as a single UT exposure, assuming multi-UT...")
        print("Telescope keywords will be computed as medians of those from UTs...")
        for k in tel_kwds.keys() :
            v=[]
            for UT in [1,2,3,4] :
                v+=[hdu[0].header.get('HIERARCH ESO TEL%d %s' %(UT,tel_kwds[k]),np.nan)]
            tel_kwds[k]='NONE'
            tel_vals[k.replace('_KEYWORD','_VALUE')]=np.median(np.array(v)[~np.isnan(v)])
    else :
        for k in tel_kwds.keys() :
            if 'HIERARCH ESO %s %s' %(tel_id,tel_kwds[k]) in hdu[0].header :
                tel_kwds[k]='ESO %s %s' %(tel_id,tel_kwds[k])
            elif 'HIERARCH ESO TEL %s' %(tel_kwds[k]) in hdu[0].header :
                tel_kwds[k]='ESO TEL %s' %(tel_kwds[k])
            else :
                tel_kwds[k]='NONE'
                print('WARNING : could not find keyword for %s in header. Using current value of %s.' %(k,k.replace('_KEYWORD','_VALUE')))

    # Barycentric radial velocity...
    QC_BERV_VALUE=0.0
    QC_BERV_KEY="NONE"
    if "HIERARCH ESO QC BERV" in hdu[0].header :
        wave_frame="VAC_RV"
        QC_BERV_KEY="ESO QC BERV"
        #QC_BERV_KEY="NONE"
        #QC_BERV_VALUE=hdu[0].header["HIERARCH ESO QC BERV"]
        if "WAVE_TV" in hdu[1].columns.names :
            clam="WAVE_TV"
            wave_frame='VAC'

    # Column names depend on the product type...
    if hdu[0].header.get('HIERARCH ESO PRO CATG','NULL') in [
            "S1D_FINAL_A",
        ] :
        clam='WAVE'
        cflux='FLUX'
        cdflux='ERR'
        #cmask='QUAL'
    elif hdu[0].header.get('HIERARCH ESO PRO CATG','NULL') in [
            "S1D_FLUXCAL_A",
            "S1D_SKYSUB_FLUXCAL_A",
        ] : #this are the values to use for ESPRESSO IDP format spectra
        cflux='flux_cal'
        cdflux='error_cal'
        #wave_frame="AIR"
 
    # Infact all ESPRESSO PL products are in an SDP-compliant Binary-table format...
    if hdu[0].header.get('HIERARCH ESO PRO CATG','NULL') in [
        "S1D_A",
        "S1D_FINAL_A",
        "S1D_FLUXCAL_A",
        "S1D_SKYSUB_A",
        "S1D_SKYSUB_FLUXCAL_A",
        "S1D_STD_A",
    ] :
        primary='FALSE'
        errorext=0
        maskext=0
        map_convolve="1,1"
        map_atmosphere="0,1"
        map_correct="0,1"
        dd=hdu[1].data[cflux]
        wl_range=np.array([np.min(hdu[1].data[clam]),np.max(hdu[1].data[clam])])
        if len(np.shape(hdu[1].data[clam])) == 1 :
            wl_resolution=np.min(hdu[1].data[clam][1:]-hdu[1].data[clam][0:-1])
        else :
            wl_resolution=np.min(hdu[1].data[clam][0][1:]-hdu[1].data[clam][0][0:-1])
    median_cont = float(np.median(dd[np.where(~np.isnan(dd))]))


    ## From Carlo Manara:
    """
    comment for the future: I would use these wl_incl for ESPRESSO
    0.6265,0.62952,0.6865,0.6938,0.7165,0.7325,0.759,0.77 since there is an
    emission line [OI] at 630 nm that is present in many (young star) spectra.
    Might be good to exclude it....
    """
    WL_inc_exc_PIX_exc={}
    inst_setup_changed = set_inst_setup(hdu[0].header) != previous_inst_setup
    if ParameterInitialization :
        ## ---------------------------------------------------------------------
        # Automatically set molecules and ranges...
        wl_span=wl_range[1]-wl_range[0]

        WL_ranges_to_include=am.auto_molecules_set_WL_ranges_to_include(
        files,
        wl_range,
        wl_span,
        wl_resolution,
        wl_scale=1./wlg,
        buff=wl_span*0.05,
        )
        WL_inc_exc_PIX_exc={
            'LIST_MOLEC': WL_ranges_to_include['list_mol'],
            'FIT_MOLEC': WL_ranges_to_include['fit_mol'],
            'REL_COL': WL_ranges_to_include['rel_col'],
            'WAVE_INCLUDE': WL_ranges_to_include['wave_include'],
            'WAVE_EXCLUDE': WL_ranges_to_include['wave_exclude'],
            'PIXEL_EXCLUDE': WL_ranges_to_include['pixel_exclude'],
        }
        ## ---------------------------------------------------------------------

    return dict(
        list(({
            # model settings...
            'COLUMN_LAMBDA': clam,
            'FLUX_UNIT': flux_unit,
            'COLUMN_FLUX': cflux,
            'COLUMN_DFLUX': cdflux,
            #Not yet, need to understand quality values... and set threshold
            #'COLUMN_MASK': cmask,
            'USE_ONLY_INPUT_PRIMARY_DATA': primary,
            'USE_DATA_EXTENSION_AS_DFLUX': errorext,
            'USE_DATA_EXTENSION_AS_MASK': maskext,
            'CONTINUUM_CONST': median_cont,
            'WAVELENGTH_FRAME': wave_frame,
            'OBS_ERF_RV_KEY': QC_BERV_KEY,
            'OBS_ERF_RV_VALUE': QC_BERV_VALUE,
            # calctrans settings...
            'MAPPING_CONVOLVE': map_convolve,
            'MAPPING_ATMOSPHERIC': map_atmosphere,
            # correct settings...
            'MAPPING_CORRECT': map_correct,
        }).items())
        +
        list(tel_kwds.items())
        +
        list(tel_vals.items())
        +
        list((WL_inc_exc_PIX_exc if inst_setup_changed else {}).items())
    )
# ------------------------------------------------------------------------------------------
